// Copyright (C) 2004 by OpenFlower Team
// http://sourceforge.net/projects/openflower/
// Licensed under the GNU GPL Version 2.

#include <iostream>

#include "AlgoAllenCahn.h"
#include "MyMacros.h"
#include "Simulation.h"
#include "MyVector.h"
#include "Variable.h"
#include "Medium.h"

#include "EqnAllenCahn.h"
#include "EqnTemperatureDiffusion.h"

#include "OpDiffusion.h"

using namespace std;



AlgoAllenCahn::AlgoAllenCahn()
  {
  nbEqn=0;
  pEqn.resize(0);
  }



AlgoAllenCahn::~AlgoAllenCahn()
  {
  }



/*!\brief Executes the Algorithm.
 *
 */
void AlgoAllenCahn::execute()
  {
  EqnAllenCahn *eq0 = dynamic_cast<EqnAllenCahn*>(pEqn[0]);
  EqnTemperatureDiffusion *eq1 = dynamic_cast<EqnTemperatureDiffusion*>(pEqn[1]);
  
  if((!eq0)||(!eq1))
    {
    ERROR_IN_FILE;
    cerr << "Specified Equations are not consistant with Algorithm"<<endl;
    cerr << "Expecting 2 Equations:"<<endl<<" First: Allen-Cahn Equation."<<endl;
    cerr << " Second: Temperature Diffusion Equation."<<endl;
    cerr << "Exiting"<<endl;
    exit(-1);
    }
  
  Variable& phi = eq0->getPresent();
  Variable& T = eq1->getPresent();
  
  Variable& phiFuture = eq0->getFuture();
  Variable& TFuture = eq1->getFuture();
  
  Variable& gamma = eq0->getGamma();
  Variable& beta = eq0->getBeta();
  Variable& lambda = eq0->getLambda();
  Variable& k = eq0->getK();

  const Variable& rLambda  = eq1->getMedium().getLambda();
  const Variable& rRho  = eq1->getMedium().getRho();
  const Variable& rCp  = eq1->getMedium().getCp();
  Variable alpha(rLambda);
  alpha/=rRho;
  alpha/=rCp;

  // w=gamma phi**2 (1-phi)**2
  // s=beta phi**2 (3-2phi)
  // dwdf = gamma (2 phi(1-3 phi (1-4 phi/6)))
  // dsdf = 6 beta phi (1-phi)
  Variable dwdf(phi);
  Variable dsdf(phi);
  Variable diff;
  OpDiffusion& op0 = dynamic_cast<OpDiffusion&>(eq0->getOperator(0));
  OpDiffusion& op1 = dynamic_cast<OpDiffusion&>(eq1->getOperator(0));

  
  double dt = pSimulation->getTimeStep();

  dwdf.setName("dWdPhi");
  
  dwdf*=(-2./3.);
  dwdf+=1.;
  dwdf*=phi;
  dwdf*=(-3.);
  dwdf+=1.;
  dwdf*=phi;
  dwdf*=2.;
  dwdf*=gamma;
  
  dsdf*=(-1.);
  dsdf+=1.;
  dsdf*=6.;
  dsdf*=phi;
  dsdf*=beta;

  Variable RHS0(T);

  op0.evaluate(lambda,phi,diff);
  RHS0.setName("RHS0");
  RHS0*=dsdf;
  RHS0*=(-1.);
  RHS0+=dwdf;
  RHS0-=diff;
  RHS0*=k;
  RHS0*=(-1.);

 
  Variable RHS1(T);

  op1.evaluate(alpha,T,diff);
  RHS1*=dsdf;
  RHS1*=(-1);
  RHS1*=RHS0;
  RHS1+=diff;
  
// Explicit Euler Time advancement
  RHS0*=dt;
  RHS1*=dt;
  RHS0+=phi;
  RHS1+=T;
  
  phiFuture = RHS0;
  TFuture = RHS1;
  }



double AlgoAllenCahn::getStabilityTimeStep()
  {
  EqnAllenCahn & eq0 = dynamic_cast<EqnAllenCahn&>(*pEqn[0]);
  EqnTemperatureDiffusion & eq1 = dynamic_cast<EqnTemperatureDiffusion&>(*pEqn[1]);
  
  double stabilityTimeStep;
  
  Variable& lambda = eq0.getLambda();
  Variable& k = eq0.getK();

  const Variable& rLambda  = eq1.getMedium().getLambda();
  const Variable& rRho  = eq1.getMedium().getRho();
  const Variable& rCp  = eq1.getMedium().getCp();
  Variable alpha(rLambda);
  alpha/=rRho;
  alpha/=rCp;

  OpDiffusion& op0 = dynamic_cast<OpDiffusion&>(eq0.getOperator(0));
  OpDiffusion& op1 = dynamic_cast<OpDiffusion&>(eq1.getOperator(0));
  
  double dt0 = op0.getStabilityTimeStep(lambda);

// (Vortexflow , 23 sept 2004) -> I am not sure about this here...
  dt0 = dt0/k.getMax();
  
  double dt1 = op1.getStabilityTimeStep(alpha);

  if(dt0<dt1) return dt0;
  else return dt1;
  }
