// Copyright (C) 2004 by OpenFlower Team
// http://sourceforge.net/projects/openflower/
// Licensed under the GNU GPL Version 2.

#include <config.h>
#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <fstream>

#include "BCHeatFlux.h"
#include "Mesh.h"
#include "Variable.h"
#include "Equation.h"
#include "Parser.h"
#include "Face.h"
#include "Element.h"
#include "Medium.h"

using namespace std;



/*!\brief Reads the Input File
 *
 * This method reads from an input stream the data file and recovers the Expression
 * giving the function to impose on the wall, eg: (z<1)*(2*x+1)+(z>1)*(2*(x-1)+0.5).
 *
 * \param inputFile A reference to the input FLW file.
 */
void BCHeatFlux::read(ifstream& inputFile)
  {
  string name;
  inputFile >> name;
  string fct_str;
  inputFile >> fct_str;
  
  bcFunction.makeTree(fct_str);
  associateFrontier(name);
  values.resizes(pFrontier->getNbFaces(),1);
  updateBCValues();
  }



/*!\brief Updates the values imposed on the Boundary
 *
 * This method updates the values specified on the boundary-condition.
 * It will be mainly called (for example when time-dependent boundary-conditions
 * are considered) before every successive time step.
 *
 */
void BCHeatFlux::updateBCValues()
  {
  double result;
  int nbFaces = pFrontier->getNbFaces();

  int faceNumber;
  MyVector<double> XYZPosition;
  XYZPosition.resize(3);
  const MyVector<Face>& faces = pMesh->getRefFaces();
  
  for(int i=0;i<nbFaces;i++)
    {
    faceNumber = pFrontier->getFace(i);
    XYZPosition = faces[faceNumber].getXYZ();

    bcFunction.setVariable("X",XYZPosition[0]);
    bcFunction.setVariable("Y",XYZPosition[1]);
    bcFunction.setVariable("Z",XYZPosition[2]);
    
    bcFunction.evaluate(result);
    values[i][0]=result;
    }
  }



/*!\brief Calculation of the scalar value at the Boundary Condition.
 *
 * Returns the value of the scalar field at the wall.
 * It is mainly used by mathematical operators (Gradient...) to have the contribution from the boundary-conditions.
 *
 * \param faceNumber The number of the face (in local ordering) from which we want the normal gradient.
 * \param comp The number of the component of the field (Not used because it deals with a scalar field).
 * \param gradElement Gradient of the field located at the adjacent element to the wall used for the corrective term.
 * \return A double value giving the scalar value at the wall given by the normal gradient calculation expression.
 */
double BCHeatFlux::getValue(const int faceNumber, const int comp, const MyVector<double>& gradElement)
  {
  double d;
  
  MyVector<double> XYZFace;
  XYZFace.resize(3);
  MyVector<double> XYZElement;
  XYZElement.resize(3);
  MyVector<double> unitSurfVect;
  unitSurfVect.resize(3);
  
  const Variable& rTemperature = pEquation->getPresent();
  const Variable& rLambda = (pEquation->getMedium()).getLambda();
  
  const MyVector<Face>& rFaces = pMesh->getRefFaces();
  const MyVector<Element>& rElements = pMesh->getRefElements();

  int faceNumberGlobal = pFrontier->getFace(faceNumber);
  int element = rFaces[faceNumberGlobal].getNeighbourElement(0);
  double temperatureElement = rTemperature.getValue(element);
  
  XYZFace = rFaces[faceNumberGlobal].getXYZ();
  XYZElement = rElements[element].getXYZ();
  unitSurfVect  = rFaces[faceNumberGlobal].getUnitSurfaceVector();

  d = (XYZFace[0]-XYZElement[0])*unitSurfVect[0]
    + (XYZFace[1]-XYZElement[1])*unitSurfVect[1]
    + (XYZFace[2]-XYZElement[2])*unitSurfVect[2];

  double correction=0;
  for(int i=0;i<3;i++)
    {
    correction+=(XYZFace[i]-XYZElement[i]-d*unitSurfVect[i])*gradElement[i];
    }
  temperatureElement+=correction;
  
  return temperatureElement+(d*values[faceNumber][0]/rLambda.getValue(element));
  }



/*!\brief Calculation of the scalar value at the Boundary Condition.
 * 
 * Returns the value of the gradient normal to the face given by the user
 * to the code, which is the value of the flux (lambda dT/dn) going IN the domain.
 * 
 *\param faceNumber The local number of the face from which to retrieve the flux.
 *\param comp The number of the component of the field (Not used because it deals with a scalar field).
 *\param gradElement Not used.
 */
double BCHeatFlux::getNormalGrad(const int faceNumber, const int comp, const MyVector<double>& gradElement)
  {
  
  // VortexFlower (08/10/2004) -> here we assume that BCHeatFlux boundary-condition is REALLY used
  // for the case of a scalar equation (temperature) because we take into account the lambda of the medium!
  const Variable& rLambda = (pEquation->getMedium()).getLambda();
  const MyVector<Face>& rFaces = pMesh->getRefFaces();

  int faceNumberGlobal = pFrontier->getFace(faceNumber);
  int element = rFaces[faceNumberGlobal].getNeighbourElement(0);
  
  return values[faceNumber][0]/rLambda.getValue(element);
  };




