// Copyright (C) 2004 by OpenFlower Team
// http://sourceforge.net/projects/openflower/
// Licensed under the GNU GPL Version 2.

#include <config.h>
#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <fstream>

#include "BCTemperatureImposed.h"
#include "Mesh.h"
#include "Variable.h"
#include "Equation.h"
#include "Parser.h"
#include "Face.h"
#include "Element.h"

using namespace std;



/*!\brief Reads the input FLW File.
 *
 * This method reads from an input stream the data file and recovers the Expression
 * giving the function to impose on the wall, eg: (z<1)*(2*x+1)+(z>1)*(2*(x-1)+0.5).
 *
 * \param inputFile A reference to the input FLW file.
 */
void BCTemperatureImposed::read(ifstream& inputFile)
  {
  string name;
  inputFile >> name;
  string fct_str;
  inputFile >> fct_str;
  
  bcFunction.makeTree(fct_str);
  associateFrontier(name);
  values.resizes(pFrontier->getNbFaces(),1);
  updateBCValues();
  }



/*!\brief Updates the values imposed on the Boundary.
 *
 * This method updates the values specified on the boundary-condition.
 * It will be mainly called (for example when time-dependent boundary-conditions
 * are considered) before every successive time step.
 *
 */
void BCTemperatureImposed::updateBCValues()
  {
  double result;
  int nbFaces = pFrontier->getNbFaces();

  int faceNumber;
  MyVector<double> XYZPosition;
  XYZPosition.resize(3);
  const MyVector<Face>& faces = pMesh->getRefFaces();
  
  for(int i=0;i<nbFaces;i++)
    {
    faceNumber = pFrontier->getFace(i);
    XYZPosition = faces[faceNumber].getXYZ();

    bcFunction.setVariable("X",XYZPosition[0]);
    bcFunction.setVariable("Y",XYZPosition[1]);
    bcFunction.setVariable("Z",XYZPosition[2]);
    
    bcFunction.evaluate(result);
    values[i][0]=result;
    }
  }



/*!\brief Calculation of the scalar value at the Boundary Condition.
 *
 * Returns the value of the scalar field at the wall.
 * It is mainly used by mathematical operators (Gradient...) to have the contribution from the boundary-conditions.
 *
 * \param faceNumber The number of the face (in local ordering) from which we want the normal gradient.
 * \param comp The number of the component of the field (Not used because it deals with a scalar field).
 * \param gradElement Gradient of the field located at the adjacent element to the wall used for the corrective term.
 */
double BCTemperatureImposed::getValue(const int faceNumber, const int comp, const MyVector<double>& gradElement)
  {
  return values[faceNumber][comp];
  }



/*!\brief Calculation of normal gradient of the scalar at the Boundary Condition.
 *
 * Returns the normal component of the gradient of the scalar field at the wall.
 * It is mainly used by mathematical operators (Diffusion...) to have the contribution from the boundary-conditions.
 *
 * \param faceNumber The number of the face (in local ordering) from which we want the normal gradient.
 * \param comp The number of the component of the field (Not used because it deals with a scalar field).
 * \param gradElement Gradient of the field located at the adjacent element to the wall used for the corrective term.
 * \return result A double value given by the normal gradient calculation expression.
 */
double BCTemperatureImposed::getNormalGrad(const int faceNumber, const int comp, const MyVector<double>& gradElement)
  {
  double d;
  double result;
  
  MyVector<double> XYZFace;
  XYZFace.resize(3);
  MyVector<double> XYZElement;
  XYZElement.resize(3);
  MyVector<double> unitSurfVect;
  unitSurfVect.resize(3);
  
  const Variable& temperature = pEquation->getPresent();
  const MyVector<Face>& faces = pMesh->getRefFaces();
  const MyVector<Element>& elements = pMesh->getRefElements();

  int faceNumberGlobal = pFrontier->getFace(faceNumber);
  int element = faces[faceNumberGlobal].getNeighbourElement(0);
  double temperatureElement = temperature.getValue(element);
  
  XYZFace = faces[faceNumberGlobal].getXYZ();
  XYZElement = elements[element].getXYZ();
  unitSurfVect  = faces[faceNumberGlobal].getUnitSurfaceVector();

  d = (XYZFace[0]-XYZElement[0])*unitSurfVect[0]
    + (XYZFace[1]-XYZElement[1])*unitSurfVect[1]
    + (XYZFace[2]-XYZElement[2])*unitSurfVect[2];

  double correction=0;
  for(int i=0;i<3;i++)
    {
    correction+=(XYZFace[i]-XYZElement[i]-d*unitSurfVect[i])*gradElement[i];
    }
  
  temperatureElement+=correction;

  result = (values[faceNumber][0]-temperatureElement)/d;

  return result;
  }
