// Copyright (C) 2004 by OpenFlower Team
// http://sourceforge.net/projects/openflower/
// Licensed under the GNU GPL Version 2.

#include<iostream>
#include<math.h>

#include<Element.h>
//#include<Face.h>
 


Element::Element()
  { 
  XYZ.resize(3);
  nbNeighbourFaces=0;
  };



Element::~Element()
  { 
  };



/*!\brief Sets the number of vertices of an element.
 *
 * This method sets the number of vertices of an element.
 *
 * \param nv Number of vertices for this element (4, 5, 6 or 8).
 */
void Element::setNbVertices(const int nv)
  {
  this->nbVertices = nv;
  vertices.resize(nv);
  }; 



/*!\brief Sets one of the vertices of an element.
 *
 * This method sets the number of one of the vertices of an element.
 *
 * \param iv Number of the vertex for this element.
 * \param num_v Global number of the vertex for this element.
 */
void Element::setVertex(const int iv, const int num_v)
  {
  assert(iv < this->nbVertices);
  vertices[iv] = num_v;
  };



/*!\brief Sets the number of faces of an element.
 *
 * This method sets the type (tetrahedral, prism, pyramid or hexahedral) of the element.
 *
 * \param _type Type of this element (4, 5, 6 or 7).
 */
void Element::setType(const int _type)
  {
  assert( ((_type>3) || (_type<8)) );
  type=_type;
  };



/*!\brief Sets the number of faces of an element.
 *
 * This method sets the number of faces of an element.
 *
 * \param n Number of faces for this element (4, 5 or 6).
 */
void Element::setNbNeighbourFaces(const int n)
  {
  nbNeighbourFaces=n;
  neighbourFaces.resize(n);
  };



/*!\brief Sets one of the neighbouring faces of an element.
 *
 * This method adds a face which is in contact to the element.
 *
 * \param n Number of the face for this element (between 0 and 6 - for an hexahedral).
 * \param num_n Global number of the face.
 */
void Element::setNeighbourFace(const int n,const int num_n)
  {
  neighbourFaces[n]=num_n;
  };



/*!\brief Calculates the properties of an element.
 *
 * This method calculates the center of gravity XYZ of an element
 * and calls the method to calculate its volume.
 *
 * \param vertexXYZ Matrix containing the XYZ coordinates of all vertices composing the element.
 */
void Element::calculateProperties(const MyMatrix<double> & vertexXYZ)
  {
  XYZ=vertexXYZ[vertices[0]];
  for (int i=1;i<nbVertices;i++)
    {
    XYZ.add(vertexXYZ[vertices[i]]);
    }
  XYZ.multiply(1./nbVertices);

  calculateVolume(vertexXYZ);
  };



/*!\brief Calculates the volume of an element.
 *
 * This method calculates the volume of an element.
 *
 * \param vectorXYZ Matrix containing the XYZ coordinates of all vertices composing the element.
 */
void Element::calculateVolume(const  MyMatrix<double> & vectorXYZ)
  {
  MyVector<double> v1,v2,v3,v4,v5,v6,v7,v8;
  v1.resize(3);v2.resize(3);v3.resize(3);v4.resize(3);
  v5.resize(3);v6.resize(3);v7.resize(3);v8.resize(3);
  v1=vectorXYZ[vertices[0]];
  v2=vectorXYZ[vertices[1]];
  v3=vectorXYZ[vertices[2]];
  v4=vectorXYZ[vertices[3]];
  // note that for quadrangle that are not planar, we calculate an average Volume  (see case 5,6,7)
  switch (type)
    {
    case 4:    volume=calculateVolumeTetrahedron(v1,v2,v3,v4); break;
    case 5:
               // calculation as multi-tetrahedrons
               v5=vectorXYZ[vertices[4]];
               v6=vectorXYZ[vertices[5]];
               v7=vectorXYZ[vertices[6]];
               v8=vectorXYZ[vertices[7]];
	       // see convention in gmsh doc p121   or in element.h
               volume=calculateVolumeTetrahedron(XYZ,v1,v2,v5);
               volume+=calculateVolumeTetrahedron(XYZ,v6,v2,v5);
               volume+=calculateVolumeTetrahedron(XYZ,v1,v2,v6);
               volume+=calculateVolumeTetrahedron(XYZ,v1,v5,v6);

               volume+=calculateVolumeTetrahedron(XYZ,v2,v1,v3);
               volume+=calculateVolumeTetrahedron(XYZ,v4,v1,v3);
               volume+=calculateVolumeTetrahedron(XYZ,v2,v1,v4);
               volume+=calculateVolumeTetrahedron(XYZ,v4,v2,v3);

               volume+=calculateVolumeTetrahedron(XYZ,v4,v1,v8);
               volume+=calculateVolumeTetrahedron(XYZ,v5,v1,v8);
               volume+=calculateVolumeTetrahedron(XYZ,v4,v1,v5);
               volume+=calculateVolumeTetrahedron(XYZ,v5,v4,v8);

               volume+=calculateVolumeTetrahedron(XYZ,v5,v8,v6);
               volume+=calculateVolumeTetrahedron(XYZ,v7,v8,v6);
               volume+=calculateVolumeTetrahedron(XYZ,v5,v7,v6);
               volume+=calculateVolumeTetrahedron(XYZ,v7,v8,v5);

               volume+=calculateVolumeTetrahedron(XYZ,v6,v2,v7);
               volume+=calculateVolumeTetrahedron(XYZ,v3,v2,v7);
               volume+=calculateVolumeTetrahedron(XYZ,v6,v2,v3);
               volume+=calculateVolumeTetrahedron(XYZ,v3,v6,v7);

               volume+=calculateVolumeTetrahedron(XYZ,v3,v4,v7);
               volume+=calculateVolumeTetrahedron(XYZ,v8,v4,v7);
               volume+=calculateVolumeTetrahedron(XYZ,v3,v8,v7);
               volume+=calculateVolumeTetrahedron(XYZ,v8,v4,v3);
               volume*=0.5;
               break;
    case 6: // Prism
               v5=vectorXYZ[vertices[4]];
               v6=vectorXYZ[vertices[5]];
// VortexFlow (25/08/2004): seems to be wrong
//                volume=calculateVolumeTetrahedron(XYZ,v1,v2,v3);
//                volume+=calculateVolumeTetrahedron(XYZ,v4,v5,v6);
//                
//                volume+=calculateVolumeTetrahedron(XYZ,v2,v3,v5);
//                volume+=calculateVolumeTetrahedron(XYZ,v6,v3,v5);
//                volume+=calculateVolumeTetrahedron(XYZ,v2,v3,v6);
//                volume+=calculateVolumeTetrahedron(XYZ,v2,v6,v5);
//                
//                volume+=calculateVolumeTetrahedron(XYZ,v2,v1,v5);
//                volume+=calculateVolumeTetrahedron(XYZ,v4,v1,v5);
//                volume+=calculateVolumeTetrahedron(XYZ,v2,v1,v4);
//                volume+=calculateVolumeTetrahedron(XYZ,v4,v2,v5);
//                
//                volume+=calculateVolumeTetrahedron(XYZ,v1,v3,v4);
//                volume+=calculateVolumeTetrahedron(XYZ,v6,v3,v4);
//                volume+=calculateVolumeTetrahedron(XYZ,v1,v3,v6);
//                volume+=calculateVolumeTetrahedron(XYZ,v6,v1,v4);
// 	          volume*=0.5;


// FIXXME (Vortexflow, 26 aug 2004) -> Works if prisms are not distorted.
// Should be checked and/or corrected to take into account
// the possible distorsion of the elements.
               volume=calculateVolumeTetrahedron(v1,v2,v3,v6);
               volume+=calculateVolumeTetrahedron(v2,v4,v5,v6);
               volume+=calculateVolumeTetrahedron(v1,v2,v4,v6);
               volume+=calculateVolumeTetrahedron(v1,v2,v3,v4);
               volume+=calculateVolumeTetrahedron(v2,v4,v5,v3);
               volume+=calculateVolumeTetrahedron(v4,v5,v6,v3);
	       volume*=0.5;
// END FIXXME               
	       
               break;
    case 7:
               v5=vectorXYZ[vertices[4]];
               volume=calculateVolumeTetrahedron(v1,v3,v2,v5);
               volume+=calculateVolumeTetrahedron(v1,v3,v4,v5);
               volume+=calculateVolumeTetrahedron(v2,v4,v3,v5);          
               volume+=calculateVolumeTetrahedron(v2,v4,v1,v5);
               volume*=0.5;
               break;
    default: cerr << "Element type not recognized!" << endl; exit(-1);
    }
  };



/*!\brief Calculates the volume of a Tetrahedral.
 *
 * This method calculates the volume of a tetrahedral given the XYZ position
 * of the four vertices composing the tetrahedral.
 *
 * \param v1 Vector containing the XYZ coordinates of vertex 1
 * \param v2 Vector containing the XYZ coordinates of vertex 2
 * \param v3 Vector containing the XYZ coordinates of vertex 3
 * \param v4 Vector containing the XYZ coordinates of vertex 4
 */
double Element::calculateVolumeTetrahedron( const MyVector<double> & v1, const MyVector<double> & v2, const MyVector<double> & v3, const MyVector<double> & v4 )
  {
  double x0,x1,x2,x3,y0,y1,y2,y3,z0,z1,z2,z3,V;
  x0=v1[0];y0=v1[1];z0=v1[2];
  x1=v2[0];y1=v2[1];z1=v2[2];
  x2=v3[0];y2=v3[1];z2=v3[2];
  x3=v4[0];y3=v4[1];z3=v4[2];

  V = fabs((x1-x0)*((y2-y0)*(z3-z0)-(y3-y0)*(z2-z0))-
           (x2-x0)*((y1-y0)*(z3-z0)-(y3-y0)*(z1-z0))+
           (x3-x0)*((y1-y0)*(z2-z0)-(y2-y0)*(z1-z0)))/6;
   return V;
  };

