// Copyright (C) 2004 by OpenFlower Team
// http://sourceforge.net/projects/openflower/
// Licensed under the GNU GPL Version 2.

#include <time.h>
#include <iostream>
#include <fstream>

#include "Mesh.h"
#include "MyMacros.h"
#include "Octree.h"
#include "Face.h"
#include "Element.h"
#include "Frontier.h"

using namespace std;

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
Mesh::Mesh()
  {
  nbEdges    = 0;
  nbFaces    = 0;
  nbElements = 0;
  nbFrontiers= 0;
  format      = "gmsh";
  MinXYZVertices.resize(3);
  MaxXYZVertices.resize(3);
  minCaractLength=1e30;
  octree = new Octree;
  }



////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
Mesh::~Mesh()
  {
  }



////////////////////////////////////////////////////////////
// Tests if mesh file exists and launches the appropriate mesh reader
////////////////////////////////////////////////////////////
void Mesh::read(ifstream& inputFile) 
  {
  string keyword;
  inputFile >> keyword;
  if(keyword != "\{")
    {
    ERROR_IN_FILE;
    cerr << "Expecting an opening \"\{\" before mesh block!" << endl;
    exit(-1);
    }
  MyVector<string> knownKeywords;
  knownKeywords.resize(3);
  knownKeywords[0]="//";
  knownKeywords[1]="file";
  knownKeywords[2]="frontiers";
  
  inputFile >> keyword;

  while (keyword!="}")
    {
    switch(knownKeywords.search(keyword))
      {
      case 0: getline(inputFile,keyword); break;
      case 1: 
             {
             inputFile >> keyword;
             file = keyword;
             cout <<         " - Mesh file          : " << file << endl;
             break;
	     }
      case 2:
              inputFile >> keyword;
              if(keyword!="{")
                {
                ERROR_IN_FILE;
              	cerr << "Expecting an opening \"\{\" before \"Frontiers\" block !" << endl;
              	exit(-1);
                }
              this->readFrontiersVirtual(inputFile);
	      break;
      case -1:
              {
              ERROR_IN_FILE;
	      cerr << keyword << " is an unknown keyword for mesh block!" << endl;
              cerr << "  Known keywords are: ";
              for (unsigned int i=0;i<knownKeywords.size();i++)
                {
		cerr << knownKeywords[i] << " ";
		}
              cerr << endl;
              exit(-1);
              break;
	      }
      default:   break;  // impossible case
      }                                                         // { of switch
      inputFile >> keyword;                            
    }                                                          // { of  while 
  
  ifstream meshFile(file.c_str());      
  if (!meshFile)
    {
    ERROR_IN_FILE;
    cerr << "Mesh file : '" << file <<"' not found !" << endl;
    exit(-1);
    }
  meshFile.close();

  readMeshVirtual();
  
  calculateFacesElementsProperties();
  calculateMinimalCharacteristicLength();
  checkMesh();

  };



////////////////////////////////////////////////////////////
// Writes the mesh in a file in openflower format
////////////////////////////////////////////////////////////
void Mesh::writeMeshFlw(ifstream& inputFile) const
  {
  string keyword;
  int i, j;
  inputFile >> keyword;
  ofstream outputFile(keyword.c_str());
  
  outputFile << nbVertices << endl;
  outputFile << nbEdges << endl;
  outputFile << nbFaces << endl;
  outputFile << nbFacesFrontiers << endl;
  outputFile << nbElements << endl;
  outputFile << nbFrontiers << endl;
  outputFile << endl;
  
  outputFile << MinXYZVertices[0]<<" "<<MinXYZVertices[1]<<" "<<MinXYZVertices[2]<<endl;
  outputFile << MaxXYZVertices[0]<<" "<<MaxXYZVertices[1]<<" "<<MaxXYZVertices[2]<<endl;
  outputFile << endl;

  for(i=0; i<nbVertices;i++)
    {
    outputFile << vertexXYZ[i][0] << " "  << vertexXYZ[i][1] << " "<< vertexXYZ[i][2] << endl;
    }
  outputFile << endl;
  
  for(i=0; i<nbEdges;i++)
    {
    outputFile << edges[i][0] << " " << edges[i][1] << endl;
    }
  outputFile << endl;

  MyVector<double> collect;
  for(i=0; i<nbFaces;i++)
    {
    outputFile << faces[i].getType()<<" "<<faces[i].getNbVertices();
    for(j=0;j<faces[i].getNbVertices();j++)
      {
      outputFile<<" "<<faces[i].getVertex(j);
      }
    
    collect = faces[i].getXYZ();
    outputFile <<" "<<collect[0]<<" "<<collect[1]<<" "<<collect[2];
    collect = faces[i].getSurfaceVector();
    outputFile <<" "<<collect[0]<<" "<<collect[1]<<" "<<collect[2];
    outputFile <<" "<<faces[i].getNeighbourElement(0)<<" "<<faces[i].getNeighbourElement(1);
    outputFile << endl;
    }
  outputFile << endl;
  
  for(i=0; i<nbElements;i++)
    {
    outputFile << elements[i].getType()<<" "<<elements[i].getNbVertices();
    for(j=0;j<elements[i].getNbVertices();j++)
      {
      outputFile<<" "<<elements[i].getVertex(j);
      }
    
    for(j=0;j<elements[i].getNbNeighbourFaces();j++)
      {
      outputFile<<" "<<elements[i].getNeighbourFace(j);
      }
    
    collect = elements[i].getXYZ();
    outputFile <<" "<<collect[0]<<" "<<collect[1]<<" "<<collect[2];
    outputFile <<" "<<elements[i].getVolume();
    outputFile <<" "<<faces[i].getNeighbourElement(0)<<" "<<faces[i].getNeighbourElement(1);
    outputFile << endl;
    }


  outputFile << endl;
  outputFile.close();
  };



//////////////////////////////////////////////////////////////
// Calculation of surfaces, volumes, and coordinates of faces/elements
////////////////////////////////////////////////////////////
void Mesh::calculateFacesElementsProperties()
  {
  int i;
  for(i=0;i<nbElements;i++) //Calculation of volume for each element
    {
    elements[i].calculateProperties(vertexXYZ);
    }

  MyVector<double> Fxyz,Exyz,S;
  Fxyz.resize(3);Exyz.resize(3);
  // calculation of orientation of the surface_vector for each element
  // (from inside to outside the element or inverse)
  for(i=0;i<nbFaces;i++)                        
    {
    Fxyz=faces[i].getXYZ();
    Exyz=elements[faces[i].getNeighbourElement(0)].getXYZ();
    S=faces[i].getSurfaceVector();
    //normal outside element for neighbour_element(0)
    if ( ((Fxyz[0]-Exyz[0])*S[0] + (Fxyz[1]-Exyz[1])*S[1] +(Fxyz[2]-Exyz[2])*S[2]) < 0)
      {
      faces[i].makeSurfaceVectorOpposite();
      }
    }
 
 // Creation of Mesh Octree for Elements
  clock_t t1=clock();
  // Creation of table containing all coordinates of the faces
  // for construction of octree of faces.
  MyMatrix<double> Tab;
  Tab.resizes(nbElements,3);
  MyVector<double> XYZ;
  XYZ.resize(3);
  for(i=0;i<nbElements;i++)
    {
    XYZ = elements[i].getXYZ();
    Tab[i][0] = XYZ[0];
    Tab[i][1] = XYZ[1];
    Tab[i][2] = XYZ[2];
    }
  octree->createOctree(MinXYZVertices,MaxXYZVertices,Tab,nbElements);
  octree->associateMesh(this);
  clock_t t2=clock();
  cout << (t2-t1)/(double)CLOCKS_PER_SEC<<" seconds elapsed for Elements Octree creation"<<endl;
  }



void Mesh::calculateMinimalCharacteristicLength()
  {
  double l=10e30;
  double Smax;
  for(int i=0;i<nbElements;i++)
    {
    Smax=0;
    for(int j=0;j<elements[i].getNbNeighbourFaces();j++)
      {
      if(faces[elements[i].getNeighbourFace(j)].getSurface()>Smax) Smax = faces[elements[i].getNeighbourFace(j)].getSurface();
      }
    l = elements[i].getVolume()/Smax;
    if(l<minCaractLength) minCaractLength = l;
    }
  }



////////////////////////////////////////////////////////////
// Checks a mesh : volume integrity, correspondance between elements...
////////////////////////////////////////////////////////////
void Mesh::checkMesh()
  {
  int i,j;
  cout << " Results of checkMesh:" << endl;
  cout << " * Smallest Characteristic Mesh Length: " << minCaractLength << " m." << endl;
  for (i=0; i<nbElements;i++) // verify that each element has the right number of neighbour faces
    {
    switch (elements[i].getType())
      {
      case 4: assert(elements[i].getNbNeighbourFaces()==4); break;
      case 5: assert(elements[i].getNbNeighbourFaces()==6); break;
      case 6: assert(elements[i].getNbNeighbourFaces()==5); break;
      case 7: assert(elements[i].getNbNeighbourFaces()==5); break;
      default:
             {
             ERROR_IN_FILE;
             cerr << "Unknown element " << i << ": type= " << elements[i].getType() << "!" << endl;
             exit(-1);
             }
      }
    }

  double Volume=0;
  for(i=0;i<nbElements;i++)
    {
    Volume+=elements[i].getVolume();
    }
  cout << " * Total volume= " << Volume << " m3." << endl;

  double WholeSurface=0;
  for (i=0;i<nbFacesFrontiers;i++)
    {
    WholeSurface+=faces[i].getSurface();
    }
  cout << " * Total frontiers surface= " << WholeSurface << " m2." << endl;


  cout << " * Frontiers surfaces: " << endl;
  int numFacesTot=0;
  double Surface = 0;
  for(i=0;i<nbFrontiers;i++)      
    {
    Surface = 0;
    const unsigned int FrontiersNf = frontiers[i].getNbFaces();  // Gets the number of faces of the Frontier
    if (FrontiersNf==0) // Checks if there is at least one surface per Frontier
      {
      ERROR_IN_FILE;
      cerr << "The frontier defined with name " << frontiers[i].getName() << " does not exist in the mesh file!" << endl;
      exit(-1);
      }
    numFacesTot+=FrontiersNf;
    for(unsigned int surf=0;surf<FrontiersNf;surf++)
      {
      Surface += faces[frontiers[i].getFace(surf)].getSurface();
      }
    cout << "   -" << frontiers[i].getName() << "= " << Surface << " m2."<<endl;
    }
  assert(numFacesTot==nbFacesFrontiers);
  }
