// Copyright (C) 2004 by OpenFlower Team
// http://sourceforge.net/projects/openflower/
// Licensed under the GNU GPL Version 2.

#include <iostream>
#include <fstream>

#include "MyMacros.h"
#include "MeshGmsh.h"
#include "Octree.h"
#include "Face.h"
#include "Element.h"
#include "Frontier.h"

using namespace std;



////////////////////////////////////////////////////////////
// Reads frontiers and domain names for gmsh files
////////////////////////////////////////////////////////////
void Mesh_gmsh::readFrontiersVirtual(ifstream& inputFile)
  {
  int num;
  unsigned int i,j;
  string keyword;

  MyVector<string> knownKeywords;
  knownKeywords.resize(1);
  knownKeywords[0]="surface";

  inputFile >> keyword;

  while (keyword!="}")
    {
    switch(knownKeywords.search(keyword))
      {
      case 0:
            {
            nbFrontiers++;
            frontiers.resize(nbFrontiers);
            inputFile >> num;
            if(inputFile.fail())
              {
              ERROR_IN_FILE;
              cerr << "A numerical value was expected after " << keyword << "!" << endl;
              exit(-1);
              }
            frontiers[nbFrontiers-1].setPhysicalNumber(num);
            inputFile >> keyword;
            frontiers[nbFrontiers-1].setName(keyword);
            break;
            }
     case -1:
            {
            ERROR_IN_FILE;
            cerr << keyword << " is an unknown keyword for Gmsh Mesh block!" << endl;
            cerr << "Known keywords are: ";
            for (i=0;i<knownKeywords.size();i++)
              {
              cerr << knownKeywords[i] << " ";
              }
            cerr << endl;
            exit(-1);
            }
      default:   break;  // Impossible case
    } // { of switch
  inputFile >> keyword;
  } // { of  while


  if(nbFrontiers<1)
    {
    ERROR_IN_FILE;
    cerr << "Problem in the definition of the simulation..."<<endl;
    cerr << "No frontier defined! "<<endl;
    exit(-1);
    }

  cout << "Finished reading "<< nbFrontiers <<" frontiers."<<endl;

  // Checks if two frontiers do not have the same name
  for(i=0;i<nbFrontiers;i++)
    {
    for(j=0;j<nbFrontiers;j++)
      {
      if(j!=i)
        {
	if(frontiers[i].getName()==frontiers[j].getName())
          {
          ERROR_IN_FILE;
          cerr << "User defined two frontiers with same name: " << frontiers[i].getName() <<endl;
          cerr << "Modify input data file. Exiting now..." <<endl;
          exit(-1);
          }
	}
      }
    }
  };





////////////////////////////////////////////////////////////
// Reader for gmsh format mesh files
////////////////////////////////////////////////////////////
void Mesh_gmsh::readMeshVirtual() 
  {
  unsigned int i,j;
  string keyword;
  MyVector<int> NumVertex; // give the number of the vertex in gmsh file (see warning below)
  MyVector<int> InvNumVertex;

  ifstream inputFile(file.c_str());

  // read the block of the vertices
  inputFile >> keyword;
  assert(keyword == "$NOD");
  inputFile >> nbVertices;
  vertexXYZ.resizes(nbVertices,3);  //   Makes the necessary resize.
  NumVertex.resize(nbVertices);
  for(i=0;i<nbVertices;i++)
    {
    inputFile >> NumVertex[i];  // WARNING: the vertices numbers in gmsh file are not equaly spaced from 1 to nb_vertices
    for(j=0;j<3;j++)
      {
      inputFile >> vertexXYZ[i][j];
      }
    }
  inputFile >> keyword;
  assert(keyword == "$ENDNOD");

  // Calculation of the inverse table of num_vertex
  int n_max= NumVertex[nbVertices-1];
  InvNumVertex.resize(n_max+1);
  for (i=0;i<n_max;i++) InvNumVertex[i]=-1;  // Initialize -1
  for (i=0;i<nbVertices;i++) InvNumVertex[NumVertex[i]]=i;


  //    In Gmsh standard, elements can be lines 1, (triangles 2, quadrangles 3,) -> surfaces
  //   (tetrahedron 4, hexahedron 5, prisms 6, pyramids 7 ) -> elements      or points 15
  // read the block of  lines / surfaces / elements
  inputFile >> keyword;
  assert(keyword == "$ELM") ;
  int nbGmshElements; // Number of Gmsh elements written in the mesh file
  inputFile >> nbGmshElements;

  // Resizes are not optimized at all here - corrrected later when nbFaces and nbElements are known
  edges.resizes(nbGmshElements,2);
  faces.resize(nbGmshElements);
  elements.resize(nbGmshElements);

  int num_tmp; 
  int gmshElementType,gmshPhysicalEntity,gmshElementaryEntity,nbVerticesGmshElement;
  MyVector<int> vertexGmshElement;   // num of the vertex of an element
  vertexGmshElement.resize(8);       //maximum of 8 vertices in an element
  MyVector<int> frontiersPhysNbList; // just to help here for searching the connectivity faces<->frontiers
  frontiersPhysNbList.resize(nbFrontiers);
  for (int i=0;i<nbFrontiers;i++)
    {
    frontiersPhysNbList[i]=frontiers[i].getPhysicalNumber();
    }
  
  for(int i=0;i<nbGmshElements;i++)
    {
    inputFile >> num_tmp; // num of the gmsh element
    inputFile >> gmshElementType;
    inputFile >> gmshPhysicalEntity;
    inputFile >> gmshElementaryEntity;
    inputFile >> nbVerticesGmshElement;
    for (int j=0;j<nbVerticesGmshElement;j++)
      {
      inputFile >> num_tmp;
      vertexGmshElement[j] = InvNumVertex[num_tmp];
      }
    if(gmshElementType == 1)                                                // for lines
      {
      edges[nbEdges][0]=vertexGmshElement[0];
      edges[nbEdges][1]=vertexGmshElement[1];
      nbEdges++;
      }
    if( (gmshElementType == 2) || ( gmshElementType == 3) )                     // for surfaces
      {
      faces[nbFaces].setType(gmshElementType);
      faces[nbFaces].setFrontierNumber(gmshPhysicalEntity);
      faces[nbFaces].setNbVertices(nbVerticesGmshElement);
      for(int j=0;j<nbVerticesGmshElement;j++)
        {
        faces[nbFaces].setVertex(j,vertexGmshElement[j]);
        }
      num_tmp=frontiersPhysNbList.search(gmshPhysicalEntity);
      if(num_tmp==-1)
        {
        ERROR_IN_FILE;
        cerr << "User forgot to define frontier number " << gmshPhysicalEntity << endl;
        exit(-1);
        }
      frontiers[num_tmp].setNbFaces(frontiers[num_tmp].getNbFaces()+1);
      frontiers[num_tmp].setFace(frontiers[num_tmp].getNbFaces()-1,nbFaces);
      nbFaces++;
      }
    if( (gmshElementType > 3) && ( gmshElementType < 8) )                     // for elements
      {
      elements[nbElements].setType(gmshElementType);
      elements[nbElements].setNbVertices(nbVerticesGmshElement);
      for(int j=0;j<nbVerticesGmshElement;j++)
        {
	elements[nbElements].setVertex(j,vertexGmshElement[j]);
	}
      nbElements++;
      }
    }

  inputFile >> keyword;
  assert(keyword == "$ENDELM") ;

  nbFacesFrontiers=nbFaces;

  cout << "       * Number of frontier faces: " << nbFacesFrontiers << endl;
  cout << "       * Number of elements      : " << nbElements << endl;

  edges.resizes(nbEdges,2);
  faces.resize(nbFaces);
  elements.resize(nbElements);
  inputFile.close();

  calculatesAllFacesGmsh();

  cout << "       * Number of faces         : " << nbFaces << endl;
  cout << "      -> Mesh file completed!"<<endl;
	
  };



////////////////////////////////////////////////////////////
// Determines all faces (not only frontiers faces
//          as in gmsh file +  neighbours elements<->faces
////////////////////////////////////////////////////////////
void Mesh_gmsh::calculatesAllFacesGmsh()
  {
  MyVector<int> v;              // numbers of the vertices of an element
  MyMatrix<int> vv;
  vector<int> sumv;              // sum of the vertices numbers of each face (only to help to find double faces)
  
  int nbFaces_tmp=nbFacesFrontiers+nbElements*4;
  faces.resize(nbFaces_tmp); // initialisation here and correction later for hexahedron, prism and pyramid
  sumv.resize(nbFaces_tmp);
  
  for(int i=0; i<nbElements;i++)  
    {                                                     // { of for int i
//  cout << "Finding all faces in domain : " << int(i*100.0/nbElements) << " % completed....\r";
    switch (elements[i].getNbVertices())
      {                                                   // { du switch
      case 4 : // tetrahedron
                v.resize(4);
                for (int j=0; j<4;j++)
                  v[j]=elements[i].getVertex(j);
                vv.resizes(4,3);
                vv[0][0]=v[0]; vv[0][1]=v[1]; vv[0][2]=v[3];
                vv[1][0]=v[0]; vv[1][1]=v[2]; vv[1][2]=v[1];
                vv[2][0]=v[0]; vv[2][1]=v[3]; vv[2][2]=v[2];
                vv[3][0]=v[1]; vv[3][1]=v[2]; vv[3][2]=v[3];
                for (int j=0; j<4;j++)             // for the four new faces
                {
                  faces[nbFaces].setType(2);               //triangle
                  faces[nbFaces].setNbVertices(3);
                  faces[nbFaces].setFrontierNumber(0);
                  faces[nbFaces].setNeighbourElement(0,i);
                  sumv[nbFaces]=0;
                  for (int k=0;k<3;k++)
                  {
                    faces[nbFaces].setVertex(k,vv[j][k]);
                    sumv[nbFaces]+=vv[j][k];
                  }
                  nbFaces++;
                }
                break;
      case 5 : // pyramid
                nbFaces_tmp++;
                faces.resize(nbFaces_tmp);
                sumv.resize(nbFaces_tmp);
                v.resize(5);
                for (int j=0; j<5;j++)                          // caculation of correspondance face <-> vertices in an order way
                  v[j]=elements[i].getVertex(j);
                vv.resizes(4,3);
                vv[0][0]=v[0]; vv[0][1]=v[1]; vv[0][2]=v[4];    // cf gmsh conventions
                vv[1][0]=v[0]; vv[1][1]=v[4]; vv[1][2]=v[3];
                vv[2][0]=v[1]; vv[2][1]=v[2]; vv[2][2]=v[4];
                vv[3][0]=v[2]; vv[3][1]=v[3]; vv[3][2]=v[4];
                for (int j=0; j<4;j++)             // for the four new triangles
                {
                  faces[nbFaces].setType(2);               //triangle
                  faces[nbFaces].setNbVertices(3);
                  faces[nbFaces].setFrontierNumber(0);
                  faces[nbFaces].setNeighbourElement(0,i);
                  sumv[nbFaces]=0;
                  for (int k=0;k<3;k++)
                  {
                    faces[nbFaces].setVertex(k,vv[j][k]);
                    sumv[nbFaces]+=vv[j][k];
                  }
                  nbFaces++;
                }
                vv.resizes(1,4);
                vv[0][0]=v[0]; vv[0][1]=v[3]; vv[0][2]=v[2]; vv[0][3]=v[1];    // cf gmsh conventions
                  faces[nbFaces].setType(3);               // one quadrangle
                  faces[nbFaces].setNbVertices(4);
                  faces[nbFaces].setFrontierNumber(0);
                  faces[nbFaces].setNeighbourElement(0,i);
                  sumv[nbFaces]=0;
                  for (int k=0;k<4;k++)
                  {
                    faces[nbFaces].setVertex(k,vv[0][k]);
                    sumv[nbFaces]+=vv[0][k];
                  }
									nbFaces++;
                break;
      case 6 : // prism
                {
                nbFaces_tmp++;
                faces.resize(nbFaces_tmp);
                sumv.resize(nbFaces_tmp);
                v.resize(6);
                for (int j=0; j<6;j++)               // caculation of correspondance face <-> vertices in an order way
                  v[j]=elements[i].getVertex(j);
                vv.resizes(2,3);
                vv[0][0]=v[0]; vv[0][1]=v[2]; vv[0][2]=v[1];            // cf gmsh conventions
                vv[1][0]=v[3]; vv[1][1]=v[4]; vv[1][2]=v[5];
                 for (int j=0; j<2;j++)             // for the two new triangles
                {
                  faces[nbFaces].setType(2);               //triangle
                  faces[nbFaces].setNbVertices(3);
                  faces[nbFaces].setFrontierNumber(0);
                  faces[nbFaces].setNeighbourElement(0,i);
                  sumv[nbFaces]=0;
                  for (int k=0;k<3;k++)
                  {
                    faces[nbFaces].setVertex(k,vv[j][k]);
                    sumv[nbFaces]+=vv[j][k];
                  }
                  nbFaces++;
                }
                vv.resizes(3,4);
                vv[0][0]=v[0]; vv[0][1]=v[1]; vv[0][2]=v[4];  vv[0][3]=v[3];           // cf gmsh conventions
                vv[1][0]=v[0]; vv[1][1]=v[3]; vv[1][2]=v[5];  vv[1][3]=v[2];
                vv[2][0]=v[1]; vv[2][1]=v[2]; vv[2][2]=v[5];  vv[2][3]=v[4];
                 for (int j=0; j<3;j++)             // for the three new quadrangles
                  {
                  faces[nbFaces].setType(3);               //quadrangle
                  faces[nbFaces].setNbVertices(4);
                  faces[nbFaces].setFrontierNumber(0);
                  faces[nbFaces].setNeighbourElement(0,i);
                  sumv[nbFaces]=0;
                  for (int k=0;k<4;k++)
                    {
                    faces[nbFaces].setVertex(k,vv[j][k]);
                    sumv[nbFaces]+=vv[j][k];
                    }
                  nbFaces++;
                  }
                break;
		}
      case 8 : // hexahedron
                nbFaces_tmp+=2;
                faces.resize(nbFaces_tmp);
                sumv.resize(nbFaces_tmp);
                v.resize(8);
                for (int j=0; j<8;j++)                // caculation of correspondance face <-> vertices in an order way
                  v[j]=elements[i].getVertex(j);
                vv.resizes(6,4);
                vv[0][0]=v[0]; vv[0][1]=v[1]; vv[0][2]=v[5];  vv[0][3]=v[4];           // cf gmsh conventions
                vv[1][0]=v[0]; vv[1][1]=v[3]; vv[1][2]=v[2];  vv[1][3]=v[1];
                vv[2][0]=v[0]; vv[2][1]=v[4]; vv[2][2]=v[7];  vv[2][3]=v[3];
                vv[3][0]=v[1]; vv[3][1]=v[2]; vv[3][2]=v[6];  vv[3][3]=v[5];
                vv[4][0]=v[2]; vv[4][1]=v[3]; vv[4][2]=v[7];  vv[4][3]=v[6];
                vv[5][0]=v[4]; vv[5][1]=v[5]; vv[5][2]=v[6];  vv[5][3]=v[7];
                 for (int j=0; j<6;j++)             // for the six new quadrangles
                {
                  faces[nbFaces].setType(3);               //quadrangle
                  faces[nbFaces].setNbVertices(4);
                  faces[nbFaces].setFrontierNumber(0);
                  faces[nbFaces].setNeighbourElement(0,i);
                  sumv[nbFaces]=0;
                  for (int k=0;k<4;k++)
                  {
                    faces[nbFaces].setVertex(k,vv[j][k]);
                    sumv[nbFaces]+=vv[j][k];
                  }
                  nbFaces++;
                }
                break;
      default :
              {
              ERROR_IN_FILE;
              cerr << "Error in read_mesh_gmsh - unknown element" << endl;
              exit(-1);
              }
      }                                                   // { du switch
    }          // } of for int i
  
  cout << "Finding all faces in domain completed...." << endl;
  assert(nbFaces==nbFaces_tmp);


  MinXYZVertices[0]=vertexXYZ.min(0);
  MinXYZVertices[1]=vertexXYZ.min(1);
  MinXYZVertices[2]=vertexXYZ.min(2);

  MaxXYZVertices[0]=vertexXYZ.max(0);
  MaxXYZVertices[1]=vertexXYZ.max(1);
  MaxXYZVertices[2]=vertexXYZ.max(2);


  for(int i=0;i<nbFacesFrontiers;i++)
    {
    sumv[i]=0;
    for(int j=0;j<faces[i].getNbVertices();j++)
      {
      sumv[i]+=faces[i].getVertex(j);
      }
    }


//////////////////////////////////
// Removing all double faces
  {
  clock_t t1=clock();
  int m;
  for(int i=0;i<nbFaces;i++)           //calculation of coordinates, surface for each face
    {
    faces[i].calculateProperties(vertexXYZ);
    }

  Octree octreeFaces;
  MyMatrix<double> Tab; 	// Creation of table containing all coordinates of the faces
				// for construction of octree of faces.
  Tab.resizes(nbFaces,3);
  MyVector<double> XYZ;
  XYZ.resize(3);
  for(int i=0;i<nbFaces;i++)
    {
    XYZ = faces[i].getXYZ();
    Tab[i][0] = XYZ[0];
    Tab[i][1] = XYZ[1];
    Tab[i][2] = XYZ[2];
    }
  octreeFaces.createOctree(MinXYZVertices,MaxXYZVertices,Tab,nbFaces);

  MyVector<int> result;
  result.resize(nbFaces);
  octreeFaces.searchDoubleFaces(result,faces,sumv);

  for(int i=0;i<nbFaces/2;i++)
    {
    m=result[i];
    faces[i].setNeighbourElement(1,faces[m].getNeighbourElement(0));
    faces[m]=faces[nbFaces-i-1];
    result[m]=result[nbFaces-i-1];
    result[result[nbFaces-i-1]] = m;
    }
  nbFaces/=2;
  faces.resize(nbFaces);

  clock_t t2=clock();
  cout << (t2-t1)/(double)CLOCKS_PER_SEC<<" seconds elapsed for finding double faces with Octree"<<endl;
  }



  // Determines neighbours faces of each element from the known neighbour elements of each face
  for(int i=0;i<nbFacesFrontiers;i++)       // For frontiers faces, only one neighbour element !
    {
    faces[i].setNeighbourElement(0,faces[i].getNeighbourElement(1));
    faces[i].setNeighbourElement(1,-1);
    }
  int p,q;
  for(int i=0;i<nbElements;i++)
    {
    elements[i].setNbNeighbourFaces(0);
    }
  for(int i=0;i<nbFacesFrontiers;i++)
    {
    p=faces[i].getNeighbourElement(0);
    q=elements[p].getNbNeighbourFaces();
    elements[p].setNbNeighbourFaces(q+1);
    elements[p].setNeighbourFace(q,i);
    }
  for(int i=nbFacesFrontiers;i<nbFaces;i++)
    {
    p=faces[i].getNeighbourElement(0);
    q=elements[p].getNbNeighbourFaces();
    elements[p].setNbNeighbourFaces(q+1);
    elements[p].setNeighbourFace(q,i);
    p=faces[i].getNeighbourElement(1);
    q=elements[p].getNbNeighbourFaces();
    elements[p].setNbNeighbourFaces(q+1);
    elements[p].setNeighbourFace(q,i);
    }
  };



////////////////////////////////////////////////////////////
// Writes a gmsh file for checking purpose only
////////////////////////////////////////////////////////////
void Mesh_gmsh::writeMeshVirtual(ifstream& inputFile) const
  {
  int i,j;
  if(format!="gmsh")
    {
    ERROR_IN_FILE;
    cerr << "Format of mesh file is not gmsh..." << endl;
    assert(format=="gmsh");
    }
  string keyword;
  inputFile >> keyword;
  ofstream outputFile(keyword.c_str());
  
  outputFile << "$NOD" << endl;
  outputFile << nbVertices << endl;
  for(i=0; i<nbVertices;i++)
    {
    outputFile << i << " " << vertexXYZ[i][0] << " " << vertexXYZ[i][1] << " "<< vertexXYZ[i][2] << endl;
    }
  outputFile << "$ENDNOD" << endl;
  outputFile << "$ELM" << endl;
  outputFile << nbEdges+nbFaces+nbElements << endl;
  for(i=0; i<nbEdges;i++)
    {
    outputFile << i << " 1 1 1 2 " << edges[i][0] << " " << edges[i][1] << endl;
    }
  for(i=0; i<nbFaces;i++)
    {
    outputFile << nbEdges+i << " " << faces[i].getType() << " 1 1 " << faces[i].getNbVertices();
    for(j=0;j<faces[i].getNbVertices();j++)
      {
      outputFile << " " << faces[i].getVertex(j) ;
      }
    outputFile << endl;
    }    
  for(i=0; i<nbElements;i++)
    {
    outputFile << nbEdges+nbFaces+i << " " << elements[i].getType() << " 1 1 " << elements[i].getNbVertices();
    for(j=0;j<elements[i].getNbVertices();j++)
      {
      outputFile << " " << elements[i].getVertex(j) ;
      }
    outputFile << endl;
    }    
  outputFile << "$ENDELM" << endl;
  outputFile << endl;
  outputFile.close();
  };

