// Copyright (C) 2004 by OpenFlower Team
// http://sourceforge.net/projects/openflower/
// Licensed under the GNU GPL Version 2.

#include <algorithm>
#include <time.h>

#include "Simulation.h"
#include "MyMacros.h"

#include "Mesh.h"
#include "Octree.h"
#include "MeshGmsh.h"

#include "Equation.h"
#include "EqnNavierStokesIncompressible.h"
#include "EqnTemperatureDiffusion.h"
#include "EqnAllenCahn.h"

#include "PostProcessing.h"

#include "Medium.h"

#include "AlgoAllenCahn.h"
#include "AlgoScalarDiff.h"

using namespace std;



Simulation::Simulation()
  {
  nbEqn=0;
  nbAlgo=0;
  nbMediums=0;
  pEqn.resize(0);
  pMediums.resize(0);
  pAlgorithms.resize(0);
  timeBegin=0.;
  time=0.;
  cfl=0.5;
  nbOfTimeStep=0;
  
  maximumTimeSteps=INT_MAX;
  };



Simulation::~Simulation()
  {  
  for(int i=0;i<pAlgorithms.size();i++)
    {
    delete pAlgorithms[i];
    }
  pAlgorithms.resize(0);

  for(int i=0;i<pEqn.size();i++)
    {
    if(pEqn[i]!=NULL) delete pEqn[i];
    }
  pEqn.resize(0);

  };



////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
void Simulation::solve()
  {
  cout <<"Solving..."<<endl;
  while(simulationIsNotOver())
    {
    PostProc->execute(time);
    
    nbOfTimeStep++;
    calculateTimeStep();
    
    cout <<nbOfTimeStep<<": time="<<time<<" , dt="<<timeStep<<" , ";
    clock_t t1=clock();
    for(int i=0;i<nbAlgo;i++)
      {
      pAlgorithms[i]->execute();
      }
    for(int i=0;i<nbEqn;i++)
      {
      pEqn[i]->advance();
      }
    
    clock_t t2=clock();
    cout << "Total CPU: "<<(t2-t1)/(double)CLOCKS_PER_SEC<<" s."<<endl;
    time+=timeStep;
    }
  finishSimulation();
  cout <<"Simulation done..."<<endl;

  };



/*!\brief Reads the input FLW file for directives.
 *
 * Reads the input FLW file and redirects to specific class readers
 * when Keyword encountered, to create all the objects specified by the user.
 *
 */
void Simulation::read()
  {
  string keyword; 
  ifstream inputFile(file.c_str()); // inputFile is the descriptor of the input data file
  if (!inputFile)
    {
    ERROR_IN_FILE;
    cerr << "Data file: '" << file <<"' not found!" << endl;
    exit(-1);
    }

  MyVector<string> knownKeywords;
  knownKeywords.resize(12);
  knownKeywords[0]="//";
  knownKeywords[1]="Mesh";
  knownKeywords[2]="write_mesh";
  knownKeywords[3]="post_processing";
  knownKeywords[4]="Equation";
  knownKeywords[5]="time_scheme";
  knownKeywords[6]="Medium";
  knownKeywords[7]="Algorithm";
  knownKeywords[8]="TimeBegin";
  knownKeywords[9]="TimeEnd";
  knownKeywords[10]="CFL";
  knownKeywords[11]="MaxTimeSteps";

  inputFile >> keyword;
               
  while(keyword!="end")
    {
    if(keyword.c_str()=="")
      {
      ERROR_IN_FILE;
      cerr << "Problem reading keyword next to " << keyword << endl;
      cerr << " verify not to forget keyword \"end\" at end of the input data file..." << endl;
      exit(-1);
      }
    switch(knownKeywords.search(keyword))
      {
      case  0:
             {
             getline(inputFile,keyword);
             break;
             }
      case  1:
             {
             readMesh(inputFile);
             break;
             }
      case  2:
             {
             pMesh->writeMeshVirtual(inputFile);
             break;
             }
      case  3:
             {
             PostProc = new PostProcessing;
             PostProc->setMesh(pMesh);
             PostProc->read(inputFile);
             break;
             }
      case  4:
             {
             addEqn(inputFile);
             break;
             }
      case  5:
             {
             ERROR_IN_FILE;
             cerr << keyword <<" is an obsolete keyword!"<<endl;
             exit(-1);
             /*readTimeScheme(inputFile);*/
             break;
             }
      case  6:
             {
             nbMediums++;
             pMediums.resize(nbMediums);
             pMediums[nbMediums-1] = new Medium;
             pMediums[nbMediums-1]->add(inputFile);
             break;
             }
      case  7:
             {
             addAlgorithm(inputFile);
             break;
             }
      case  8:
             {
             inputFile >> timeBegin;
	     time = timeBegin;
             break;
             }
      case  9:
             {
             inputFile >> timeEnd;
             break;
             }
      case  10:
             {
             inputFile >> cfl;
             break;
             }
      case  11:
             {
             inputFile >> maximumTimeSteps;
             break;
             }
      case -1:
             {
             ERROR_IN_FILE;
             cerr << keyword << " is an unknown keyword!" << endl;
             cerr << "  Known keywords are: ";
             for (unsigned int i=1;i<knownKeywords.size();i++)
               {
               cerr << knownKeywords[i] << " " ;
               }
             cerr << endl;
             exit(-1);
             break;
             }
      default: break;  // impossible case
      }
    inputFile >> keyword;
    }

  inputFile.close();

  cout << "End of simulation data file reading" << endl;

  };



/*!\brief Reads and adds a Mesh to be used for the Simulation.
 *
 * It reads in the input FLW data file the keyword designating the mesh format
 * to be used by the Simulation, creates it and initializes the attribute of the Simulation: pMesh.
 * Eventually it calls the specific read method of the Mesh.
 *
 * \param inputFile A reference to the input FLW data file.
 */
void Simulation::readMesh(ifstream& inputFile)
  {
  string keyword;
  MyVector<string> knownKeywords;
  knownKeywords.resize(1);
  knownKeywords[0]="gmsh";

  inputFile >> keyword;
  
  switch(knownKeywords.search(keyword))
    {
    case 0:
           {
           pMesh = new Mesh_gmsh ;
           pMesh->setFormat(keyword);
           break;
           }
    case -1:
           {
           ERROR_IN_FILE;
           cerr << "Specified mesh format: "<<keyword<< " is not available at the moment..." << endl;
           cerr << "  Possible mesh formats are:";
           for (unsigned int i=0;i<knownKeywords.size();i++)
             {
             cerr << " " << knownKeywords[i];
             }
           cerr << endl;
           exit(-1);
           break;
           }
    default: break;
    }

  pMesh->read(inputFile);
  }



/*!\brief Adds an Equation to the Simulation.
 *
 * It adds an Equation to the Simulation corresponding to the keyword containing
 * the name of the equation.
 * Eventually it calls the specific read method of the Equation.
 *
 * \param inputFile A reference to the input FLW data file.
 */
void Simulation::addEqn(ifstream& inputFile)
  {
  string keyword;
  MyVector<string> knownKeywords;
  knownKeywords.resize(3);
  knownKeywords[0]="scalar_diffusion";
  knownKeywords[1]="navier_stokes_incompressible";
  knownKeywords[2]="allen_cahn";

  inputFile >> keyword;
  
  switch(knownKeywords.search(keyword))
    {
    case 0:
           {
           nbEqn++;
           pEqn.resize(nbEqn);
           pEqn[nbEqn-1] = new EqnTemperatureDiffusion;
           break;
           }
    case 1:
           {
           nbEqn++;
           pEqn.resize(nbEqn);
           pEqn[nbEqn-1] = new EqnNavierStokesIncompressible;
           break;
           }
    case 2:
           {
           nbEqn++;
           pEqn.resize(nbEqn);
           pEqn[nbEqn-1] = new EqnAllenCahn;
           break;
           }
    case -1:
           {
           ERROR_IN_FILE;
           cerr << "Specified equation: "<<keyword<< " is not available at the moment..." << endl;
           cerr << "  Possible equations are:";
           for (unsigned int i=0;i<knownKeywords.size();i++)
             {
             cerr << " " << knownKeywords[i];
             }	     
           cerr << endl;
           exit(-1);
           break;
           }
    default: break;
    }

  pEqn[nbEqn-1]->associateMesh(pMesh);
  pEqn[nbEqn-1]->associateSimulation(this);
  pEqn[nbEqn-1]->read(inputFile);
  }



void Simulation::addAlgorithm(ifstream& inputFile)
  {
  string keyword;
  MyVector<string> possibleAlgorithms;
  possibleAlgorithms.resize(2);
  possibleAlgorithms[0]="AlgoAllenCahn";
  possibleAlgorithms[1]="AlgoScalarDiff";

  inputFile >> keyword;
  nbAlgo++;
  pAlgorithms.resize(nbAlgo);

  switch(possibleAlgorithms.search(keyword))
    {
    case 0:
          {
          pAlgorithms[nbAlgo-1] = new AlgoAllenCahn;
          break;
          }
    case 1:
          {
          pAlgorithms[nbAlgo-1] = new AlgoScalarDiff;
          break;
          }
    case -1:
           {
           ERROR_IN_FILE;
           cerr << "Specified Algorithm: "<<keyword<< " is not available..." << endl;
           cerr << "  Possible Algorithms are:";
           for (unsigned int i=0;i<possibleAlgorithms.size();i++)
             {
             cerr << " " << possibleAlgorithms[i];
             }
           cerr << endl;
           exit(-1);
           break;
           }
    default: break;
    }
  pAlgorithms[nbAlgo-1]->read(inputFile);
  pAlgorithms[nbAlgo-1]->associateSimulation(this);
  }



/*!\brief Prepares and checks all the attributes and sub-attributes of the Simulation.
 *
 *
 */
void Simulation::prepare()
  {
  
  int i, j, k;
  // Step 1:
  // Initializes all the Algorithms with their Equations.
  string sEq;
  int found;
  for(i=0;i<nbAlgo;i++)
    {
    for(j=0;j<pAlgorithms[i]->getNbEqn();j++)
      {
      found=0;
      sEq=pAlgorithms[i]->getLabelOf(j);
      for(k=0;k<nbEqn;k++)
        {
        if(sEq==pEqn[k]->getLabel())
          {
          found=1;
          pAlgorithms[i]->associateEquation(j,pEqn[k]);
          break;
          }
        }
      if(!found)
        {
        ERROR_IN_FILE;
        cerr << " Warning: Equation with label "<< sEq<<" was not defined in the input FLW file!"<<endl;
        exit(-1);
        }
      }
    }
  
  // Step 2:
  // Associates the Mediums with their Equations.
  for(k=0;k<nbEqn;k++)
    {
    found=0;
    for(i=0;i<nbMediums;i++)
      {
      if(pMediums[i]->getLabel()==pEqn[k]->getLabelofMedium())
        {
        found=1;
        pEqn[k]->associateMedium(pMediums[i]);
        break;
        }
      }
    if(!found)
      {
      ERROR_IN_FILE;
      cerr << " Warning: Medium with label "<< pEqn[k]->getLabelofMedium() <<endl;
      cerr << " defined in Equation "<< pEqn[k]->getLabel() << endl;
      cerr << " was not defined as a Medium in the input FLW file!"<<endl;
      exit(-1);
      }
    }

  // Step 3:
  // We fill the Map containing the label and the pointer for each Medium and Equation Variables.
  for(k=0;k<nbEqn;k++)
    {
    pEqn[k]->addsItsVariables4PostProcessing(mapNameVariable);
    }
  for(k=0;k<nbMediums;k++)
    {
    pMediums[k]->addsItsVariables4PostProcessing(mapNameVariable);
    }
  
// FIXXXME (Vortexflow, 23 sept 2004) -> Only constant or space dependant values are possible at the moment
// No dependance between other variables are possible.
  const int nbElements = pMesh->getNbElements();
  for(i=0;i<nbMediums;i++)
    {
    pMediums[i]->setNbValuesOfProperties(nbElements);
    pMediums[i]->updateProperties();
    }
// END FIXXXME

  associateVariables2PostProcessings();
  initializesProbes2PosProcessings();
  
  }



/*!\brief Find the reference of the variable in equations that are called
 * in the PostProcGlobal block.
 */
void Simulation::associateVariables2PostProcessings()
  {
  MyVector<PostGlobal> & postGlobalList = PostProc->getPostGlobalList();
  for(int i=0;i<PostProc->getnbPostGlobal();i++)
    {
    const Variable * var = getVariableFromMap(postGlobalList[i].getFieldName());
    postGlobalList[i].setRefVariable(var);
    postGlobalList[i].createOutputFile();
    }
  MyVector<PostProbe> & postProbeList = PostProc->getPostProbeList();
  for(int i=0;i<PostProc->getnbPostProbe();i++)
    {
    const Variable * var = getVariableFromMap(postProbeList[i].getFieldName());
    postProbeList[i].setRefVariable(var);
    }
  }


  
/*!\brief Initializes the specific data of Probes Post-processing.
 * 
 * Find the element in which the probe exists, gives the number of this element
 * as well as the new exact coordinates (of the element) of the probe position to the Probe.
 */
void Simulation::initializesProbes2PosProcessings()
  {
  MyVector<PostProbe> & postProbeList = PostProc->getPostProbeList();
  MyVector<double> position;
  MyVector<double> elementPosition;
  int elementNumber;
  //position.resize(3);
  //elementPosition.resize(3);
  
  for(int i=0;i<PostProc->getnbPostProbe();i++)
    {
    position = postProbeList[i].getProbePosition();
    (pMesh->myOctree()).searchElementWithCoordinates(position, elementNumber, elementPosition);
    postProbeList[i].setProbePosition(elementPosition);
    postProbeList[i].setElementNumberOfProbe(elementNumber);
    postProbeList[i].createOutputFile(i);
    }
  }



void Simulation::calculateTimeStep()
  {
  timeStep = 1e30;
  double dt;
  for(int i=0;i<nbAlgo;i++)
    {
    dt = pAlgorithms[i]->getStabilityTimeStep();
    if(dt<timeStep) timeStep = dt;
    }
  timeStep*=cfl;
  if((time+timeStep)>timeEnd)
    {
    timeStep = timeEnd-time;
    }
  }



bool Simulation::simulationIsNotOver()
  {
  if(nbOfTimeStep>=maximumTimeSteps)
    {
    cout << "Maximum number of time steps reached..."<<endl;
    return 0;
    }
  else if(time>=timeEnd)
    {
    cout << "Final time reached..."<<endl;
    return 0;
    }
  else return 1;
  }

  

const Variable* Simulation::getVariableFromMap(string& label)
  {
  NameAndVariablePointerMap::iterator position;
  position = mapNameVariable.find(label);
  if(position != mapNameVariable.end())
    {
    return position->second;
    }
  else
    {
    ERROR_IN_FILE;
    cerr << "Variable \""<< label<<"\" was not found in the list of possible"<<endl
         << "variables available for PostProcessing:"<<endl;
    for(position = mapNameVariable.begin(); position != mapNameVariable.end(); ++position)
      {
      cerr << "  " << position->first << endl;
      }
    cerr <<"Exiting..."<<endl;
    exit(-1);
    }
  }


  
void Simulation::finishSimulation()
  {
  PostProc->finish(time);
  
  // saves all the data of an Equation, basically the values of the main unknown...
  for(int i=0;i<nbEqn;i++)
    {
    pEqn[i]->finish();
    }
  }

