// Copyright (C) 2004 by OpenFlower Team
// http://sourceforge.net/projects/openflower/
// Licensed under the GNU GPL Version 2.

#include <iostream>
//#include <time.h>

#include "MyMacros.h"
#include "OpDiffusion.h"
#include "OpInterpolation.h"
#include "OpGradient.h"
#include "Element.h"
#include "Face.h"
#include "MyVector.h"
#include "BoundaryCondition.h"
#include "Equation.h"
#include "Mesh.h"

using namespace std;



OpDiffusion::OpDiffusion()
  {
  unitSurfVect.resize(3);
  position.resize(3);
  positionElement.resize(3);
  XYZ_P.resize(3);
  XYZ_E.resize(3);
  XYZ_f.resize(3);
  itIsThefirstTimeStep=1;
  
  opInterp = new OpInterpolation;
  }



/*!\brief Diffusion operator evaluation of a Variable field.
 *
 * Calculates the diffusion of a Variable field (scalar or vectorial only)
 * located at cells centers.
 *
 * \param field A reference to the Variable Field expressed at cells centers.
 * \param diffusivity A reference to the diffusivity associated to this Variable Field and expressed at the Faces.
 * \param result A reference to the returned Variable Field containing the Diffusion at all elements.
 */
void OpDiffusion::evaluate(const Variable& diffusivity, const Variable& field, Variable& result)
  {
//  clock_t t1=clock();

  const int nbFaces = pMesh->getNbFaces();
  const int nbFacesFrontiers = pMesh->getNbFacesFrontiers();
  const int nbElements = pMesh->getNbElements();

  result.setName("Diff_"+field.getName());
  result.setLocation("cell");
  result.resize(nbElements);

  const MyVector<Face>& faces = pMesh->getRefFaces();
  const MyVector<Element>& elements = pMesh->getRefElements();

  double val;
  int nbr0, nbr1;
  int i, j, k, m;
  int nbDims0 = field.getNbDimensions0();
  int nbDims1 = field.getNbDimensions1();

  MyVector<double> gradElement;
  gradElement.resize(3);

  if(field.getLocation()=="face")
    {
    ERROR_IN_FILE;
    cerr << "Field passed as argument of OpDiffusion::evaluate must be located at Cells centers."<<endl;
    cerr << "Field name is: "<<field.getName()<<endl;
    exit(-1);
    }
  
  if(itIsThefirstTimeStep)
    {
    opInterp->associateMesh(pMesh);
    opInterp->initialize();
    }

// The gradient operator
  OpGradient opGrad;
  opGrad.associateMesh(pMesh);


  assert(nbDims1==1); // Diffusion can only be calculated on a scalar or vectorial field, not a tensor (ie. nbDims1=1 ! )
//  clock_t t2=clock();

// Step 1.
// Calculates the Field on the Faces.
  Variable fieldOnFaces(field.getName(),nbFaces,nbDims0,nbDims1);
  opInterp->evaluateLinearly(field,fieldOnFaces);
  
  for(k=0;k<3;k++)
    {
    gradElement[k]=0.;
    }

  const MyVector<Frontier>& front=pMesh->getRefFrontiers();
  const MyVector<BoundaryCondition*> bcList=pEquation->getRefBC();
  int nbBC = bcList.size();
  for(i=0;i<nbBC;i++)
    {
    int nbFacesOfBC = bcList[i]->getNbFaces();
    int nbFirstFaceOfBC = bcList[i]->getNbFirstFace();
    for(j=0;j<nbFacesOfBC;j++)
      {
      for(m=0;m<nbDims0;m++)
        {
        fieldOnFaces.setValue(bcList[i]->getValue(j,m,gradElement),nbFirstFaceOfBC + j, m);
        }
      }
    }
//  clock_t t3=clock();

// Step 2.
// We calculate a first guess for the gradient Field at the center of the Elements
// gradient is the estimation at the previous time step.
// For the first time step, we use the estimation at present time step.
  if(itIsThefirstTimeStep)
    {
    opGrad.evaluate(fieldOnFaces,gradient);
    itIsThefirstTimeStep=0;
    }
//  clock_t t4=clock();

// Step 3.
// Now we calculate it on the Faces
  Variable gradientOnFaces(gradient.getName(),nbFaces,nbDims0,3);
  clock_t t45=clock();
  opInterp->evaluateWithMean(gradient,gradientOnFaces);
//  clock_t t5=clock();

// Step 4.
// We correct the scalar estimation on the center of the Faces
  Variable correctedFieldOnFaces(field.getName(),nbFaces,nbDims0);
  opInterp->evaluateWithGradient(fieldOnFaces,gradientOnFaces,correctedFieldOnFaces);
  for(i=0;i<nbBC;i++)
    {
    int nbFacesOfBC = bcList[i]->getNbFaces();
    int nbFirstFaceOfBC = bcList[i]->getNbFirstFace();
    for(j=0;j<nbFacesOfBC;j++)
      {
      nbr0 = faces[nbFirstFaceOfBC + j].getNeighbourElement(0);

      for(m=0;m<nbDims0;m++)
        {
        for(k=0;k<3;k++)
          {
          gradElement[k]=gradient.getValue(nbr0, m, k);
          }
        correctedFieldOnFaces.setValue(bcList[i]->getValue(j,m,gradElement), nbFirstFaceOfBC + j, m);
        }
      }
    }

//    clock_t t6=clock();

// Step 5.
// We recalculate the gradient Field at the center of the Elements
// thanks to the final guess of the Field at the center of the Faces
  opGrad.evaluate(correctedFieldOnFaces,gradient);
//  clock_t t7=clock();

// Step 6.
// We finally calculate the gradient normal to the face passing through the center of the face
// thanks to the estimation of the Field at orthogonal points E' and P'.
// This way we can calculate the diffusive flux accross face i.
  double EEp;
  double valEp;
  double valPp;
  double dEpPp;
  double lambda;
  double surface;

  Variable diffusivityOnFaces(diffusivity.getName(),nbFaces);
  opInterp->evaluateWithMean(diffusivity,diffusivityOnFaces);

  for(i=nbFacesFrontiers;i<nbFaces;i++)
    {
    nbr0  = faces[i].getNeighbourElement(0);
    nbr1  = faces[i].getNeighbourElement(1);
    unitSurfVect  = faces[i].getUnitSurfaceVector();
    surface = faces[i].getSurface();
    XYZ_f = faces[i].getXYZ();

    XYZ_E = elements[nbr0].getXYZ();
    XYZ_P = elements[nbr1].getXYZ();

    for(m=0;m<nbDims0;m++)
      {
      // lambda is defined as: lambda = -Ef.n
      lambda = -(XYZ_f[0]-XYZ_E[0])*unitSurfVect[0] - (XYZ_f[1]-XYZ_E[1])*unitSurfVect[1] - (XYZ_f[2]-XYZ_E[2])*unitSurfVect[2];
      // Calculation of the Field at E':
      valEp = field.getValue(nbr0,m);
      for(j=0;j<3;j++)
        {
        EEp = XYZ_f[j]-XYZ_E[j] + lambda*unitSurfVect[j];
        valEp += EEp * gradient.getValue(nbr0,m,j);
        }

      // Calculation of the Field at P':
      lambda = -(XYZ_f[0]-XYZ_P[0])*unitSurfVect[0] - (XYZ_f[1]-XYZ_P[1])*unitSurfVect[1] - (XYZ_f[2]-XYZ_P[2])*unitSurfVect[2];
      valPp = field.getValue(nbr1,m);
      for(j=0;j<3;j++)
        {
        EEp = XYZ_f[j]-XYZ_P[j] + lambda*unitSurfVect[j];
        valPp += EEp * gradient.getValue(nbr1,m,j);
        }

      dEpPp = (XYZ_P[0]-XYZ_E[0])*unitSurfVect[0]
            + (XYZ_P[1]-XYZ_E[1])*unitSurfVect[1]
            + (XYZ_P[2]-XYZ_E[2])*unitSurfVect[2];

      val = (valPp-valEp)*surface/dEpPp;

      result.addValue( diffusivityOnFaces.getValue(i)*val/elements[nbr0].getVolume(), nbr0, m);
      result.addValue(-diffusivityOnFaces.getValue(i)*val/elements[nbr1].getVolume(), nbr1, m);
      }
    }

  double normalGrad;
  int nbFacesOfBC;
  int nbFirstFaceOfBC;
  for(i=0;i<nbBC;i++)
    {
    nbFacesOfBC = bcList[i]->getNbFaces();
    nbFirstFaceOfBC = bcList[i]->getNbFirstFace();
    for(j=0;j<nbFacesOfBC;j++)
      {
      nbr0    = faces[nbFirstFaceOfBC + j].getNeighbourElement(0);
      surface = faces[nbFirstFaceOfBC + j].getSurface();
      for(m=0;m<nbDims0;m++)
        {
        for(k=0;k<3;k++)
          {
          gradElement[k]=gradient.getValue(nbr0, m, k);
          }
        normalGrad = bcList[i]->getNormalGrad(j, m, gradElement);
// FIXXME (Vortexflow, 23 sept 2004) -> For the moment, the diffusivity used at the boundary face
// takes into account the diffusivity in the neighboring element, which is not exactly the case...
// Here it should take the diffusivity AT the face!
        result.addValue(diffusivity.getValue(nbr0)*normalGrad*surface/elements[nbr0].getVolume(), nbr0, m);
// END FIXXME
        }
      }
    }
//  clock_t t9=clock();

/*
  cerr << "####### Total CPUtime for OpDiff: "<<long(t9-t1)/(double)CLOCKS_PER_SEC<<" s. ########"<<endl;
  cerr << " - Step1 (init all):            "<<(t2-t1)/(double)CLOCKS_PER_SEC<<" s. ["<< (t2-t1)*100./(t9-t1) << "%]"<<endl;
  cerr << " - Step2 (calc field on faces): "<<(t3-t2)/(double)CLOCKS_PER_SEC<<" s. ["<< (t3-t2)*100./(t9-t1) << "%]"<<endl;
  cerr << " - Step3 (calc grad):           "<<(t4-t3)/(double)CLOCKS_PER_SEC<<" s. ["<< (t4-t3)*100./(t9-t1) << "%]"<<endl;
  cerr << " - Step4 (create vargradonface):"<<(t45-t4)/(double)CLOCKS_PER_SEC<<" s. ["<< (t45-t4)*100./(t9-t1) << "%]"<<endl;
  cerr << " - Step4 (interp with mean):    "<<(t5-t45)/(double)CLOCKS_PER_SEC<<" s. ["<< (t5-t45)*100./(t9-t1) << "%]"<<endl;
  cerr << " - Step5 (calc field on faces): "<<(t6-t5)/(double)CLOCKS_PER_SEC<<" s. ["<< (t6-t5)*100./(t9-t1) << "%]"<<endl;
  cerr << " - Step5 (calc grad):           "<<(t7-t6)/(double)CLOCKS_PER_SEC<<" s. ["<< (t7-t6)*100./(t9-t1) << "%]"<<endl;
  cerr << " - Step6:                       "<<(t9-t7)/(double)CLOCKS_PER_SEC<<" s. ["<< (t9-t7)*100./(t9-t1) << "%]"<<endl;
 */
  }



double OpDiffusion::getStabilityTimeStep(const Variable& diffusivity)
  {
  const int nbElements = pMesh->getNbElements();
  const MyVector<Face>& rFaces = pMesh->getRefFaces();
  const MyVector<Element>& rElements = pMesh->getRefElements();
  double timeStep=1e30;
  double l;
  double Smax;
  double dt;
  for(int i=0;i<nbElements;i++)
    {
    Smax=0;
    for(int j=0;j<rElements[i].getNbNeighbourFaces();j++)
      {
      if(rFaces[rElements[i].getNeighbourFace(j)].getSurface()>Smax) Smax = rFaces[rElements[i].getNeighbourFace(j)].getSurface();
      }
    l = rElements[i].getVolume()/Smax;
    dt = l*l/(2*diffusivity.getValue(i));
    if(dt<timeStep) timeStep = dt;
    }
  return timeStep;
  }
