// Copyright (C) 2004 by OpenFlower Team
// http://sourceforge.net/projects/openflower/
// Licensed under the GNU GPL Version 2.

#include <iostream>
//#include <time.h>

#include "MyMacros.h"
#include "Operator.h"
#include "OpGradient.h"
#include "Mesh.h"
#include "MyVector.h"
#include "Element.h"
#include "Face.h"

using namespace std;



OpGradient::OpGradient()
{
}



/*!\brief Gradient operator evaluation of a Variable field.
 *
 * Calculates gradient of a Variable with contribution of all internal Faces.
 * It can calculate the gradient only on scalar or vectorial fields
 * (not tensors, ie. only when nbDimensions1=0 !).
 *
 * \param field A reference to the Variable Field expressed at cells centers.
 * \param result A reference to the returned Variable Field containing the Gradient at all elements.
 */
void OpGradient::evaluate(const Variable& field, Variable& result)
  {
//  clock_t t1=clock();
  
  unsigned int i, j, k;
  int nbFaces = pMesh->getNbFaces();
  int nbFacesBoundary = pMesh->getNbFacesFrontiers();
  int nbElements = pMesh->getNbElements();
  int nbDimensions0 = field.getNbDimensions0();
  int nbDimensions1 = field.getNbDimensions1();
  
//  clock_t t2=clock();

  assert(nbDimensions1==1); // Gradient operator can only calculate gradient of a scalar or vectorial field (ie nbDimensions1=1 ! )
  result.setName("Grad_"+field.getName());
  result.setLocation("cell");
  result.resize(nbElements,nbDimensions0,3); // Resizes the vector AND initializes it to 0.
  
//  clock_t t3=clock();

  const MyVector<Face> & faces = pMesh->getRefFaces();
  const MyVector<Element> & elements = pMesh->getRefElements();
  
//  clock_t t4=clock();

  double val;
  double vol0, vol1;
  double fieldVal;
  MyVector<double> surf;
  surf.resize(3);
  int nbr0, nbr1;

//  clock_t t5=clock();
  
  if(field.getLocation()=="cell")
    {
    ERROR_IN_FILE;
    cerr << "Field passed as argument of OpGradient::evaluate must be located at the Faces."<<endl;
    cerr << "Field name is: "<<field.getName()<<endl;
    exit(-1);
    }
  
//  clock_t t6=clock();
    
  for(i=0;i<nbFacesBoundary;i++)
    {
    nbr0 = faces[i].getNeighbourElement(0);
    vol0 = elements[nbr0].getVolume();
    
    surf = faces[i].getSurfaceVector();
    for(j=0;j<nbDimensions0;j++)
      {
      fieldVal = field.getValue(i,j);
      for(k=0;k<3;k++)
        {
        val = fieldVal*surf[k]/vol0;
        result.addValue(val,nbr0,j,k);
        }
      }
    }

//  clock_t t7=clock();
    
  for(i=nbFacesBoundary;i<nbFaces;i++)
    {
    nbr0 = faces[i].getNeighbourElement(0);
    vol0 = elements[nbr0].getVolume();
    
    nbr1 = faces[i].getNeighbourElement(1);
    vol1 = elements[nbr1].getVolume();
    
    surf = faces[i].getSurfaceVector();
    for(j=0;j<nbDimensions0;j++)
      {
      fieldVal = field.getValue(i,j);
      for(k=0;k<3;k++)
        {
        val = fieldVal*surf[k];
        result.addValue( val/vol0,nbr0,j,k);
        result.addValue(-val/vol1,nbr1,j,k);
        }
      }
    }
//  clock_t t8=clock();
    
/*  cout <<endl;
  cout << "####### Total CPUtime for OpGradient: "<<long(t8-t1)/(double)CLOCKS_PER_SEC<<" s. ########"<<endl;
  cout << " - Step1 :"<<(t2-t1)/(double)CLOCKS_PER_SEC<<" s. ["<< (t2-t1)*100./(t8-t1) << "%]"<<endl;
  cout << " - Step2 :"<<(t3-t2)/(double)CLOCKS_PER_SEC<<" s. ["<< (t3-t2)*100./(t8-t1) << "%]"<<endl;
  cout << " - Step3 :"<<(t4-t3)/(double)CLOCKS_PER_SEC<<" s. ["<< (t4-t3)*100./(t8-t1) << "%]"<<endl;
  cout << " - Step4 :"<<(t5-t4)/(double)CLOCKS_PER_SEC<<" s. ["<< (t5-t4)*100./(t8-t1) << "%]"<<endl;
  cout << " - Step5 :"<<(t6-t5)/(double)CLOCKS_PER_SEC<<" s. ["<< (t6-t5)*100./(t8-t1) << "%]"<<endl;
  cout << " - Step6 :"<<(t7-t6)/(double)CLOCKS_PER_SEC<<" s. ["<< (t7-t6)*100./(t8-t1) << "%]"<<endl;
  cout << " - Step7 :"<<(t8-t7)/(double)CLOCKS_PER_SEC<<" s. ["<< (t8-t7)*100./(t8-t1) << "%]"<<endl;
*/
  };



