// Copyright (C) 2004 by OpenFlower Team
// http://sourceforge.net/projects/openflower/
// Licensed under the GNU GPL Version 2.

#include <iostream>
#include <math.h>

#include "OpInterpolation.h"
#include "MyMacros.h"
#include "Element.h"
#include "Face.h"
#include "Mesh.h"
#include "Variable.h"

using namespace std;


OpInterpolation::OpInterpolation()
{
}



////////////////////////////////////////////////////////////
// Initializes the operator, ie it calculates the gamma interpolation factor
// when interpolating linearly.
// No Returns :
//  -> Creates the vector containing the values of gamma
////////////////////////////////////////////////////////////
void OpInterpolation::initialize()
  {
  const int nbFaces = pMesh->getNbFaces();
  const int nbFacesFrontiers = pMesh->getNbFacesFrontiers();

  const MyVector<Face>& faces = pMesh->getRefFaces();
  const MyVector<Element>& elements = pMesh->getRefElements();

  int i, j;
  int nbr0, nbr1;
  double dEf;
  double dfP;
  double lambda;
  // Corresponds to the XYZ position of the element denoted by nbr1
  MyVector<double> XYZ_P;
  XYZ_P.resize(3);
  // Corresponds to the XYZ position of the present element denoted by nbr0
  MyVector<double> XYZ_E;
  XYZ_E.resize(3);
  // XYZ position of the face center
  MyVector<double> XYZ_f;
  XYZ_f.resize(3);
  MyVector<double> surf;
  surf.resize(3);

  gamma.resize(nbFaces);
  dIf.resizes(nbFaces,3);


// Calculates the gamma linear interpolation factor for each internal face
  for(i=nbFacesFrontiers;i<nbFaces;i++)
    {
    nbr0 = faces[i].getNeighbourElement(0);
    nbr1 = faces[i].getNeighbourElement(1);
    XYZ_E = elements[nbr0].getXYZ();
    XYZ_P = elements[nbr1].getXYZ();
    XYZ_f = faces[i].getXYZ();

    dEf = sqrt((XYZ_E[0]-XYZ_f[0])*(XYZ_E[0]-XYZ_f[0])
    	      + (XYZ_E[1]-XYZ_f[1])*(XYZ_E[1]-XYZ_f[1])
	      + (XYZ_E[2]-XYZ_f[2])*(XYZ_E[2]-XYZ_f[2])
               );
    dfP = sqrt((XYZ_P[0]-XYZ_f[0])*(XYZ_P[0]-XYZ_f[0])
              + (XYZ_P[1]-XYZ_f[1])*(XYZ_P[1]-XYZ_f[1])
              + (XYZ_P[2]-XYZ_f[2])*(XYZ_P[2]-XYZ_f[2])
               );

    gamma[i] = dfP/(dEf+dfP);
    }

// Calculates the vector between the intersection point I and the face center f.
  for(i=0;i<nbFaces;i++)
    {
    nbr1 = faces[i].getNeighbourElement(1);
    if(nbr1==-1)
      {
      for(j=0;j<3;j++)
        {
        dIf[i][j] = 0;
        }
      }
    else
      {
      surf = faces[i].getSurfaceVector();
      nbr0 = faces[i].getNeighbourElement(0);
      XYZ_E = elements[nbr0].getXYZ();
      XYZ_P = elements[nbr1].getXYZ();
      XYZ_f = faces[i].getXYZ();
      // lambda is defined as : lambda = Ef.n / EP.n
      lambda = ((XYZ_f[0]-XYZ_E[0])*surf[0] + (XYZ_f[1]-XYZ_E[1])*surf[1] + (XYZ_f[2]-XYZ_E[2])*surf[2])
              /((XYZ_P[0]-XYZ_E[0])*surf[0] + (XYZ_P[1]-XYZ_E[1])*surf[1] + (XYZ_P[2]-XYZ_E[2])*surf[2]);

      for(j=0;j<3;j++)
        {
        dIf[i][j] = XYZ_f[j]-XYZ_E[j] - lambda*(XYZ_P[j]-XYZ_E[j]);
        }
      }
    }
  };



////////////////////////////////////////////////////////////
// Interpolates a Field from a first estimation at the face intercept I
// with a corrective term with the help of a Gradient Field expressed at point I
// Parameter :
//  - the Field expressed on the face intercept I
//  - the Gradient Field on the face intercept I
// Returns :
//  - the interpolated Field expressed at the face center f
////////////////////////////////////////////////////////////
void OpInterpolation::evaluateWithGradient(const Variable& field, const Variable& gradient, Variable& result)
  {
  const int nbFaces = pMesh->getNbFaces();
  const int nbFacesFrontiers = pMesh->getNbFacesFrontiers();
  const int nbElements = pMesh->getNbElements();

  const MyVector<Face>& faces = pMesh->getRefFaces();
  
  int nbDimensions0 = field.getNbDimensions0();
  int nbDimensions1 = field.getNbDimensions1();

  int nbDimensions0Grad = gradient.getNbDimensions0();
  int nbDimensions1Grad = gradient.getNbDimensions1();
  
  assert(nbDimensions1==1); // Interpolation with Gradient correction can only be done on a scalar and a vector (ie. nbDimensions1=1)
  assert(nbDimensions1Grad==3); // Interpolation with Gradient correction can only be done if the gradient is really a gradient with 3 space derivatives.
  assert(nbDimensions0==nbDimensions0Grad);

  int i, j, k;
  int nbr0, nbr1;
  double estimation;

  result.setNbDimensions0(nbDimensions0);
  result.setNbDimensions1(nbDimensions1);
  result.setName(field.getName());
  if((field.getLocation()=="face")&&(gradient.getLocation()=="face"))
    {
    result.setLocation("face");
    result.resize(nbFaces,nbDimensions0,nbDimensions1);

    for(i=nbFacesFrontiers;i<nbFaces;i++)
      {
      nbr0 = faces[i].getNeighbourElement(0);
      nbr1 = faces[i].getNeighbourElement(1);
      for(j=0;j<nbDimensions0;j++)
        {
        estimation = field.getValue(i,j);
        for(k=0;k<3;k++)
          {
          estimation+= gradient.getValue(i,j,k)*dIf[i][k];
          }
        result.setValue(estimation, i,j);
        }
      }
    }
  else
    {
    ERROR_IN_FILE;
    cerr << "OpInterpolation::EvaluateWithGradient for faces needs to be developped! "<<endl <<"Exiting"<<endl;
    assert(0);
    exit(-1);
    }
  };



////////////////////////////////////////////////////////////
// Interpolates a Variable Field from one location (eg center of elements)
// to another one (eg face center) linearly.
// Parameter :
//  - the Field expressed at cells centers or on the faces
// Returns :
//  - the same Field expressed at the other location (faces or elements)
////////////////////////////////////////////////////////////
void OpInterpolation::evaluateWithMean(const Variable& field, Variable& result)
  {
  const int nbFaces = pMesh->getNbFaces();
  const int nbFacesFrontiers = pMesh->getNbFacesFrontiers();
  const int nbElements = pMesh->getNbElements();

  int nbDimensions0 = field.getNbDimensions0();
  int nbDimensions1 = field.getNbDimensions1();

  const MyVector<Face>& faces = pMesh->getRefFaces();
  
  int i, j, k;
  int nbr0, nbr1;

  if(field.getLocation()=="cell")
    {
    result.setLocation("face");
    result.setName(field.getName());
    result.resize(nbFaces,nbDimensions0,nbDimensions1);

    for(i=nbFacesFrontiers;i<nbFaces;i++)
      {
      nbr1 = faces[i].getNeighbourElement(1);
      nbr0 = faces[i].getNeighbourElement(0);
      for(j=0;j<nbDimensions0;j++)
        {
        for(k=0;k<nbDimensions1;k++)
          {
          result.setValue(0.5*(field.getValue(nbr0,j,k)+field.getValue(nbr1,j,k)), i,j,k);
          }
        }
      }
    }
  else if(field.getLocation()=="face")
    {
    ERROR_IN_FILE
    cerr << "OpInterpolation::EvaluateWithMean for faces needs to be developped! "<<endl <<"Exiting"<<endl;
    assert(0);
    exit(-1);
    result.setLocation("cell");
    result.resize(nbFaces,nbDimensions0,nbDimensions1);
    }
  };



////////////////////////////////////////////////////////////
// Interpolates a Variable Field from one location (eg center of elements)
// to another one (eg face center) linearly.
// Parameter :
//  - the Field expressed at cells centers or on the faces
// Returns :
//  - the same Field expressed at the other location (faces or elements)
////////////////////////////////////////////////////////////
void OpInterpolation::evaluateLinearly(const Variable& field, Variable& result)
  {
  const int nbFaces = pMesh->getNbFaces();
  const int nbFacesFrontiers = pMesh->getNbFacesFrontiers();
  const int nbElements = pMesh->getNbElements();

  int nbDimensions0 = field.getNbDimensions0();
  int nbDimensions1 = field.getNbDimensions1();

  const MyVector<Face>& faces = pMesh->getRefFaces();

  int i, j, k;
  int nbr0, nbr1;

  result.setNbDimensions0(nbDimensions0);
  result.setNbDimensions1(nbDimensions1);
  result.setName(field.getName());
  if(field.getLocation()=="cell")
    {
    result.setLocation("face");
    result.resize(nbFaces,nbDimensions0,nbDimensions1);
    for(i=nbFacesFrontiers;i<nbFaces;i++)
      {
      nbr1 = faces[i].getNeighbourElement(1);
      nbr0 = faces[i].getNeighbourElement(0);
      for(j=0;j<nbDimensions0;j++)
        {
        for(k=0;k<nbDimensions1;k++)
          {
          result.setValue(gamma[i]*field.getValue(nbr0,j,k)+(1.-gamma[i])*field.getValue(nbr1,j,k), i,j,k);
          }
	}
      }
    }
  else if(field.getLocation()=="face")
    {
    ERROR_IN_FILE;
    cerr << "OpInterpolation::EvaluateLinearly for faces needs to be developped! "<<endl <<"Exiting"<<endl;
    assert(0);
    exit(-1);
    result.setLocation("cell");
    result.resize(nbFaces,nbDimensions0,nbDimensions1);
    }
  };
